@@ -24,6 +24,8 @@ import sys
 
 import psutil
 
+import datetime as dt
+
 #================ OPERATIONS ON NAMES, INFO ON COLS ROWS
 #------------------------------------------------------------------------
 
@@ -97,9 +99,9 @@ def get_number_of_columns(filename):
     ps.wait()
     ncolumns=len(output.decode("utf8").strip().split())
     if ncolumns==5:
-      print(f"D... {ncolumns} cols ... 2020+ version with pile-up info")
+      print(f"D... {ncolumns} cols ... 2020+ version with pile-up info (1/0)")
     elif ncolumns==4:
-        print(f"D... {ncolumns} cols  ... 2018 version without pileup column")
+        print(f"D... {ncolumns} cols  ... 2018 version without extras column, pileup (3/0)")
     else:
         print(f"X... BAD NUMBER OF COLUMNS {ncolumns}")
         sys.exit()
@@ -138,64 +140,74 @@ def fill_hist( his, narr):
     FILL histogram from np.histogram data.
     simple way bin by bin
     """
-    print(f"D... filling TH  bin by bin ({narr.shape[0]}) bins")
+    #print(f"D... filling TH  bin by bin ({narr.shape[0]}) bins")
     if len(narr.shape) == 1: # 1D
         for i in range(narr.shape[0]):
-            his.SetBinContent( i, narr[i] )
+            #if narr.shape[0]<6:
+            #    print(i, narr[i], "  ", end=";")
+            # we start with 0, bin starts with 1 in ROOT
+            his.SetBinContent( i+1, narr[i] )
 
     elif narr.shape[1] == 2: #  2D not tested
         for i in range(narr.shape[0]):
             for j in range(narr.shape[1]):
-                his.SetBinContent( i, j, narr[i][j] )
-    print("D... filling TH DONE")
+                his.SetBinContent( i+1, j+1, narr[i][j] )
+    #print("D... filling TH DONE")
 
 
 #------------------------------------------------------------------------
-def column_to_histo( dfcol , binmax = 32*1024, savename = "", hname = "h1", writeondisk = False):
+def column_to_histo( dfcol , binmax = 32*1024, himax=32*1024, savename = "", hname = "h1", writeondisk = False , writetxt = True):
     """
     send df["E"] for example
     - converts to numpy array; creates HIST;
     - saves txt
     - updates allhist.root too
     """
-    print("D... creating histo .... column to_numpy")
+    print("D... ========= creating histo ======= column to_numpy; bins=", binmax)
     narr = dfcol.to_numpy()
-    print("D... len=", len(narr), narr.dtype, ".. now to histo")
-    his,bins = np.histogram(narr,bins=binmax,range=(0.0,binmax) )
-    print("D... histo:", his)
+    print("D... len=", len(narr), narr.dtype, "...  ndarray to np histo")
+    #print(dfcol)
+    #print(narr)
+    his,bins = np.histogram(narr,bins=binmax,range=(0.0,himax) )
+    #    print("D... histo:", his)
+    #    print("D... bins :", bins)
+    del narr
 
     #
-    basename = os.path.basename(filename) # title without path...
+    basename = os.path.basename(savename) # title without path...
 
-    th1f = ROOT.TH1F(hname, basename, binmax, 0 , binmax)
+    th1f = ROOT.TH1F(hname, basename, binmax, 0 , himax)
     #for i in range(len(his)):
     #    th1f.SetBinContent( i, his[i] )
     #print("D... narr=",narr)
-    print(narr.shape  ,len(narr.shape) )
-    print(narr.shape[0] )
+
+    #print(narr.shape  ,len(narr.shape) )
+    #print(narr.shape[0] )
     fill_hist( th1f,  his )
-    th1f.Print()
+    #th1f.Print()
 
     if (savename!="") and  writeondisk:
-        print(f"D... creating text spectrum = {savename}")
-        np.savetxt(savename, his, fmt="%d")
+        if writetxt:
+            print(f"D...          text spectrum = {savename}")
+            np.savetxt(savename, his, fmt="%d")
         # rootname = os.path.splitext(savename)[0]+".root"
         rootname = os.path.dirname(savename)
         if len(rootname)>3:
             rootname+="/"
         rootname+="all_histograms.root"
-        print("D... creating root spectum at", rootname )
+        print("D...          root spectum at", rootname, "name=", hname )
         f = ROOT.TFile(rootname,"UPDATE")
-        print(f"D... saving {th1f.GetName()} into {rootname}")
+        # print(f"D...          {th1f.GetName():13s} @ {rootname}")
         th1f.Write()
         f.Close()
+    endblock(nomem=True)
     return his
 
 
 #------------------------------------------------------------------------
 
 def save_to_tree(df2, filename):
-    print("i... creating and writing tree, expect 25% of the original size")
+    print("i... ===============creating and writing tree, expect 25% of the original size")
     # convert to dict with numpy arrays
     #data = {key: df[key].values for key in ['time', 'E', 'x']}
     data = {key: df2[key].values for key in df2.columns }
@@ -203,36 +215,95 @@ def save_to_tree(df2, filename):
     rdf = ROOT.RDF.MakeNumpyDataFrame( data )
     print(rdf)
     outfile =  os.path.splitext(filename)[0] +"_tree.root"
+    print("D...  snaphost rdf as ttree to ", outfile)
+
+    start = dt.datetime.now()
     rdf.Snapshot("df",outfile)
+    meastime = (dt.datetime.now() - start).total_seconds()
+    print(f"D... speed = {get_number_of_lines(filename)/meastime/1000/1000:.3f} Mrows/sec")
+
     return
 
 #========================== READ; ANALYSIS AND DETECTION ===================
 
-def pd_link_next(df):
+# def pd_link_next(df):
+#     print( freemem() )
+#  #
+#     print(f"D... broadening table - dtus")
+#     df['dtus'] = (df.time.shift(-1)-df.time)*1000*1000
+#     df.fillna(0)
+#     df['dtus'] = df['dtus'].astype('float32') #
+#     print(df)
+#     print( freemem() )
+
+
+#     print(f"D... broadening table - next_E")
+#     df['next_E'] = (df.E.shift(-1))
+#     df.fillna(0, inplace=True)
+#     df['next_E'] = df['next_E'].astype('int32') #
+#     print( freemem() )
+
+#     # print(df) # seems to stay the same
+#     print(df)
+#     print(df.dtypes)
+#     print( freemem() )
+
+#     return df
+
+def endblock(nomem=False):
+    if not nomem:
+        print( freemem() )
+    print("D... ","_"*50)
+
+#------------------------------------------------------------------------
+def select_channels(df, channels, delete_original = False):
+    """
+    select channels bu not as view! COPY and DELETE
+    """
+    print("D... selecting channels:", channels)
     print( freemem() )
- #
-    print(f"D... broadening table - dtus")
+    # CHANNEL (i hope it is sorted)
+    if delete_original:
+        df1 = df.loc[ df["ch"].isin( channels ) ].copy()
+        del df
+    else:
+        df1 = df.loc[ df["ch"].isin( channels ) ]
+    df1.reset_index(inplace=True, drop=True)
+    #del df
+    if len(df1) ==0 :
+        print("X... no data for channels",channels)
+        sys.exit(1)
+    endblock()
+    return df1
+
+
+def enhance_by_dtus_and_next_E(df):
+    print("D... adding dtus and next_E columns")
+    print( freemem() )
+
+    print(f"D... broadening table - dt")
     df['dtus'] = (df.time.shift(-1)-df.time)*1000*1000
     df.fillna(0)
     df['dtus'] = df['dtus'].astype('float32') #
-    print(df)
-    print( freemem() )
-
+    print("D... range of erlang values : ",df['dtus'].min(), df['dtus'].max() )
 
     print(f"D... broadening table - next_E")
     df['next_E'] = (df.E.shift(-1))
     df.fillna(0, inplace=True)
     df['next_E'] = df['next_E'].astype('int32') #
-    print( freemem() )
 
-    # print(df) # seems to stay the same
+    if 'extras' in df.columns:
+        print(f"D... broadening table - prev_dtus SATU")
+        df['prev_satu'] = (df.time - df.time.shift(+1))*1000*1000
+        df['prev_satu'] = np.where((df['extras']&1)==1, df['prev_satu'] , 0)
+        df.fillna(0)
+        df['prev_satu'] = df['prev_satu'].astype('float32') #
+
     print(df)
-    print(df.dtypes)
-    print( freemem() )
-
+    endblock()
     return df
 
-#------------------------------------------------------------------------
+#----------------------------------------------------------------------
 def pd_detect_zeroes(df, channel,  TIMEWIN_US = 25  ):
     """
    1-get the channel only
@@ -242,80 +313,45 @@ def pd_detect_zeroes(df, channel,  TIMEWIN_US = 25  ):
       2nd phase - SZ and DZ - make one event from these PILEUPs
       3rd phase - create
     """
-
-    print( freemem() )
-    # CHANNEL (i hope it is sorted)
-    df = df.loc[ df["ch"] == channel ]
-    if len(df) ==0 :
-        return 1,0
-    #
-    print(f"D... broadening table - dt")
-    df['dtus'] = (df.time.shift(-1)-df.time)*1000*1000
-    df.fillna(0)
-    df['dtus'] = df['dtus'].astype('float32') #
-    print(df)
+    print( "D.. detecting zeroes")
     print( freemem() )
 
-
-    print(f"D... broadening table - dt")
-    df['next_E'] = (df.E.shift(-1))
-    df.fillna(0, inplace=True)
-    df['next_E'] = df['next_E'].astype('int32') #
-    print( freemem() )
-
-    # print(df) # seems to stay the same
-    print(df)
-    print(df.dtypes)
-
-
-
-    # ---fast way and memory saving to create histogram
-    binmax = 100
-    narr = df['dtus'].to_numpy()
-    print("D... len=", len(narr), narr.dtype, ".. now to histo")
-    his,bins = np.histogram(narr,bins=binmax,range=(0.0,binmax) )
-
-
-    # #print(f"D... plotting DOUBLE the window RANGE (range={TIMEWIN_US})")
-    # #plt.plot( his, '.' )
-    # #plt.xlabel('dt [us]')
-    # #plt.ylabel(f"ch={channel}")
-    # #plt.savefig("z_time_distr.png")
-    # #sys.exit(0)
-    # #--- df is prepared
-
-
-    print("D... searching for correlations")
-    df_dz = df.loc[ ((df.E==0) & (df.next_E==0)) & (df.dtus<TIMEWIN_US) ]
-    dzeroes = len(df_dz)
+    # this oneliner doesnt occupy memory ---------------
+    # print("D... searching for correlations")
+    dzeroes = len( df.loc[ ((df.E==0) & (df.next_E==0)) & (df.dtus<TIMEWIN_US) ])
+    #df_dz = df.loc[ ((df.E==0) & (df.next_E==0)) & (df.dtus<TIMEWIN_US) ]
+    #dzeroes = len(df_dz)
     print(freemem() )
 
-    df_sz = df.loc[ ((df.E!=0) & (df.next_E==0)) & (df.dtus<TIMEWIN_US) ]
-    szeroes = len(df_sz)
+    szeroes = len( df.loc[ ((df.E!=0) & (df.next_E==0)) & (df.dtus<TIMEWIN_US) ] )
+    #szeroes = len(df_sz)
     print(freemem() )
 
-    df_iz = df.loc[ ((df.E==0) & (df.next_E!=0)) & (df.dtus<TIMEWIN_US) ]
-    izeroes = len(df_iz)
+    izeroes = len( df.loc[ ((df.E==0) & (df.next_E!=0)) & (df.dtus<TIMEWIN_US) ] )
+    #izeroes = len(df_iz)
     print(freemem() )
 
 
     print(f"D... ---------------------- START of zero detection ch={channel}")
-    df_z = df[ (df.E==0) & (df.ch==channel)]
-    zeroes = len(df_z)
+    zeroes = len( df.loc[ (df.E==0) & (df.ch==channel)] )
+    zeroes2 = len( df[ (df.E==0) & (df.ch==channel)] )
+    #zeroes = len(df_z)
     print(freemem() )
 
 
     print(f"D... total  zeroes (chan={channel}) = {zeroes} which is  {zeroes/len(df)*100:5.2f}%")
+    print(f"D... total  zeroes (chan={channel}) = {zeroes2} which is  {zeroes2/len(df)*100:5.2f}%")
     print(f"D... double zeroes (chan={channel}) = {dzeroes} which is  {dzeroes/len(df)*100:5.2f}%")
-    print(df_dz)
+    #print(df_dz)
     print(f"D... single zeroes (chan={channel}) = {szeroes} which is  {szeroes/len(df)*100:5.2f}%")
-    print(df_sz)
+    #print(df_sz)
     print(f"D... ilogic zeroes (chan={channel}) = {izeroes} which is  {izeroes/len(df)*100:5.2f}%")
-    print(df_iz)
+    #print(df_iz)
     print("D... ______________________ end of zero detection")
 
 
     # returns length for now
+    endblock()
     return zeroes,dzeroes
 
 #------------------------------------------------------------------------
@@ -323,26 +359,54 @@ def pd_detect_zeroes(df, channel,  TIMEWIN_US = 25  ):
 def pd_read_table(filename, sort = True):
     """
     READ THE ASC TABLE:   two possibilities:  4 columns OR 5 columns
+EARLIER:
+         Events[ch][ev].TimeTag, ENERGY,(Events[ch][ev].Energy)>>MAXNBITS, ch
+         MAXNBITS 15
+Feb 2020+:
+       Events[ch][ev].TimeTag, ENERGY,  (Events[ch][ev].Energy)>>MAXNBITS, ch, Events[ch][ev].Extras
+       MAXNBITS 15
+    >>t,e,PU (1/0) (or 3/0 old)  ,  channel,   Extras(2020+ only)
+
+
+https://www.npl.washington.edu/TRIMS/sites/sand.npl.washington.edu.TRIMS/files/manuals-documentation/CAENDigitizer_SW_User_Manual_rel.5.pdf
+
+EXTRAS: bit[0] = DEAD_TIME. This is set to 1 when a dead time occurred before this event. The dead time can be due to
+either a signal saturation or a full memory status. Check Fig. B.4 and Fig. B.5 for more details
+bit[1] = ROLL_OVER. Identify a trigger time stamp roll-over that occurred before this event
+bit[2] = TT_RESET. Identify a trigger time stamp reset forced from external signals in S-IN (GPI for Desktop)
+bit[3] = FAKE_EVENT. This is a fake event (which does not correspond to any physical event) that identifies a
+time stamp roll-over. The roll-over can be due to an external or internal reset. The user can set bit[25]
+= 1 of register 0x1n80 to enable the fake-event saving in case of reset from S-IN, and bit[26] = 1 of
+register 0x1n80 to enable the fake-event saving in case of internal roll-over. In the first case the event
+will have both bit[3] and bit[2] set to 1, while in the second case the event will have both bit[3] and
+bit[1] set to 1.
     """
+    start = dt.datetime.now()
     df = pd.read_table( filename,
-                        names=['time','E','x','ch','y'],
+                        names=['time','E','pu','ch','extras'],
                         sep = "\s+",
                         comment="#")
 
+    meastime = (dt.datetime.now() - start).total_seconds()
+    print(f"D... speed = {get_number_of_lines(filename)/meastime/1000/1000:.3f} Mrows/sec")
+
+
+    print("D... lasttime ",df.iloc[-1]["time"])
     print(freemem() )
     # drop if 4 column
     df = df.dropna(axis='columns', how='all')
 
 
     df['time'] = df['time']/1e+8 # this is correct for 10ns step ADC
+    print("D... lasttime ",df.iloc[-1]["time"])
 
     #print(df.dtypes)
     df['time'] = df['time'].astype('float64')
     df['E']    = df['E'].astype('int32')
-    df['x']    = df['x'].astype('int32')
+    df['pu']    = df['pu'].astype('int32')
     df['ch']   = df['ch'].astype('int32')
-    if 'y' in df:
-        df['y'] = df['y'].astype('int32')
+    if 'extras' in df:
+        df['extras'] = df['extras'].astype('int32')
     print("D...    finale types ************************************")
     print(df.dtypes)
     # print(freemem())
@@ -363,6 +427,8 @@ def pd_read_table(filename, sort = True):
         #print(df)
         #print( freemem() )
 
+    print("D... lasttime ",df.iloc[-1]["time"])
+    endblock()
     return df
 
 
