@@ -25,7 +25,7 @@ from shutil import copyfile
 import sys
 
 import ROOT
-
+from tabulate import tabulate
 
 #>>> import pandas as pd
 #>>> import matplotlib.pyplot as plt
@@ -185,10 +185,39 @@ def fastread(filename, x=0, y=1, batch = 0, read_last=0, df = None, plot = True)
     return df2
 
 
+def coincidences(filename, ch0, ch1,  tree = False):
+    """
+    """
+    print( general.freemem() )
 
+    #*************
+    nlines   = general.get_number_of_lines(filename)
+    ncolumns = general.get_number_of_columns(filename)
+    print(f"i... reading the table of {nlines} lines ... ({nlines/1e+6:.1f} milion lines)")
+    print(f"i... reading the table of {ncolumns} columns ... ")
+
+
+    #*************
+    df = general.pd_read_table(filename, sort = True)
+    #print(df.dtypes)
+    print(df)
+    chan_available = df['ch'].unique()
+    print("D... channels available:", chan_available )
+    print("D... x        present:",df['pu'].unique() )
+    if 'extras' in df:
+        print("D... y        present:",df['extras'].unique() )
+    print("D... Emin            :",df['E'].min() )
+    print("D... Emax            :",df['E'].max() )
+
+    if (ch0 in chan_available) or (ch1 in chan_available):
+        print(f"X... channels {ch0} OR {ch1} not in the file !!")
+        sys.exit(1)
+
+
+    return
 
 #=============================== MAIN1 EVA ==================
-def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
+def eva_cut_time(filename,  chan, od=0, do=9999999 ,   tree = False):
     """
     EVA: use: read_vme_offline cut1 filename_with_asc  60 120
     """
@@ -196,6 +225,9 @@ def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
     # do = 999999
 
 
+
+
+    print("D... expect 30Mevents processed per minute at best")
     print( general.freemem() )
 
     #*************
@@ -226,17 +258,21 @@ def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
     df = general.pd_read_table(filename, sort = True)
     #print(df.dtypes)
     print(df)
+
+    print("+"*50,"TOTALS")
     chan_available = df['ch'].unique()
-    print("D... channels present:", chan_available)
-    if chan in chan_available:
-        print(f"X... no events in channel {chan}")
-        return
-    print("D... x        present:",df['x'].unique() )
-    if 'y' in df:
-        print("D... y        present:",df['y'].unique() )
+    print("D... channels available:", chan_available, "   chan selected:", chan)
+    print(f"D... pu        present: {df['pu'].unique()} ... 1 or 0 (3v0 before 2020)" )
+    if 'extras' in df:
+        print(f"D... extras    present: {df['extras'].unique()}... 1..satur,3..roll,4..reset,8..fake" )
+        print(f"D... total saturations: ",len( df.loc[ (df['extras']&1)==1]  ) )
     print("D... Emin            :",df['E'].min() )
     print("D... Emax            :",df['E'].max() )
 
+    if not(chan in chan_available):
+        print(f"X... no events in channel {chan}")
+        return
+
 
 
     if (od!=0) and (do<9999999):
@@ -245,7 +281,7 @@ def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
         print(f"D...  selecting events for  channel {chan} from {od}s to {do}s")
         print(f"D...  selecting events for  channel {chan} from {od}s to {do}s")
         print()
-        df1 = df[ (df.ch==chan)&(df.time>od)&(df.time<do)  ].copy() # copy ELSE warning....
+        df1 = df[ (df.time>od)&(df.time<do)  ].copy() # copy ELSE warning....
         df1.reset_index(inplace=True, drop=True)
         df1.fillna(0,inplace=True)
 
@@ -253,12 +289,97 @@ def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
         if len(df1)==0:
             print("D... no data for channel {chan}")
             sys.exit(0)
+
     else:
         df1 = df
 
+    #---------------------------------------------------DF1 ---------------------------
+    #************* reduce the DF
+    df1 = general.select_channels(df1, [chan] , delete_original = True)
+
+
+
+    print("+"*50,"CHANNEL",chan)
+    chan_available = df1['ch'].unique()
+    print("D... channels available:", chan_available, "   chan selected:", chan)
+    print(f"D... pu        present: {df1['pu'].unique()}... 1 or 0 (3v0 before 2020)" )
+    if ncolumns == 5:
+        pileups_ch = len(df1.loc[ df1['pu']==1])
+    else:
+        pileups_ch = len(df1.loc[ df1['pu']==3])
+
+    print(f"D... pu        present: {df1['pu'].unique()}... 1 or 0 (3v0 before 2020)" )
+    if 'extras' in df:
+        print(f"D... extras    present: {df1['extras'].unique()}... 1..satur,3..roll,4..reset,8..fake" )
+        print(f"D... number of saturations: ",len( df1.loc[ (df1['extras']&1)==1]  ) )
+    print("D... Emin            :",df1['E'].min() )
+    print("D... Emax            :",df1['E'].max() )
+
+
 
     #*************  dt us    and next_E
-    df = general.pd_link_next(df)
+    df1 = general.enhance_by_dtus_and_next_E(df1)
+    satu_n_ch = len(df1.loc[ (df1['extras']&1)!=0] )
+    print("D... SATURATIONS\n",df1.loc[ (df1['extras']&1)!=0] )
+    satu_DT_ch = df1.loc[ (df1['extras']&1)!=0]["prev_satu"].sum()
+    print(f"D... SATURATIONS {satu_n_ch} total: {satu_DT_ch/1e+6} sec.")
+
+
+    # not final
+    stat_columns = ['ch','rate','realT','liveT','deadT','satT','min_dtus','Nzero','Npu','Ntot','Nsat','DT%','DT+S%']
+    df_stat = pd.DataFrame( np.nan, index=[0], columns = stat_columns )
+    df_stat['ch'] = chan
+    df_stat['Nsat'] = satu_n_ch
+    df_stat['satT'] = satu_DT_ch/1e+6
+    df_stat['min_dtus'] = df1.loc[df1['dtus']!=0]['dtus'].min()
+    df_stat['Npu'] = pileups_ch
+
+
+    #------------- histogram with time erlang---------------------------------
+    bname = os.path.splitext(filename)[0]  #basename
+    # hname = bname.split("_")[-1]  # last thing should be a comment
+    #*************
+    hname = general.generate_hname(filename, chan)+"_erlang"
+    outfile = bname+"erlang_ch"+str(chan)+".txt"
+    #*************
+    his = general.column_to_histo(df1['dtus'], binmax=1000, himax=1+int(df1["dtus"].max()),savename = outfile, hname = hname, writeondisk = True, writetxt = False)
+    hname = general.generate_hname(filename, chan)+"_erlang2"
+    outfile = bname+"erlang2_ch"+str(chan)+".txt"
+    #*************
+    his = general.column_to_histo(df1['dtus'], binmax=1000, himax=100,savename = outfile, hname = hname, writeondisk = True, writetxt = False)
+
+
+
+    #--------------- histogram zeros in time
+    hname = general.generate_hname(filename, chan)+"_zerotime"
+    outfile = bname+"zerotime_ch"+str(chan)+".txt"
+    #*************
+    his = general.column_to_histo(  df1.loc[ (df1.E==0)  ]["time"]  , binmax=1+int(df1.iloc[-1]["time"]), himax=1+int(df1.iloc[-1]["time"]), savename = outfile, hname = hname, writeondisk = True, writetxt = False)
+
+
+    #--------------- histogram nonzeroes in time
+    hname = general.generate_hname(filename, chan)+"_nzerotime"
+    outfile = bname+"nzerotime_ch"+str(chan)+".txt"
+    #*************
+    his = general.column_to_histo(  df1["time"].loc[ (df1.E!=0)  ]  , binmax=1+int(df1.iloc[-1]["time"]), himax=1+int(df1.iloc[-1]["time"]), savename = outfile, hname = hname, writeondisk = True, writetxt = False)
+
+
+    #--------------- histogram saturations distribution in time
+    hname = general.generate_hname(filename, chan)+"_saturations"
+    outfile = bname+"saturations_ch"+str(chan)+".txt"
+    #*************
+    his = general.column_to_histo(  df1["time"].loc[ (df1.prev_satu!=0)  ]  , binmax=1+int(df1.iloc[-1]["time"]), himax=1+int(df1.iloc[-1]["time"]), savename = outfile, hname = hname, writeondisk = True, writetxt = False)
+
+
+    #--------------- histogram saturations length
+    hname = general.generate_hname(filename, chan)+"_satlength"
+    outfile = bname+"satlength_ch"+str(chan)+".txt"
+    #*************
+    his = general.column_to_histo(  df1["prev_satu"].loc[ (df1.prev_satu!=0)  ]  , binmax=100, himax=1+int(df1["prev_satu"].max()), savename = outfile, hname = hname, writeondisk = True, writetxt = False)
+
+
+
+
 
     #*************   one channel; time span
     len_dfzero,_ = general.pd_detect_zeroes(df1, chan) # df1[ (df1.E==0) ]
@@ -285,14 +406,21 @@ def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
 #    print(f"i... DT %   == {deadtpr:.2f}")
     # print(f"i... events == {fev} ... {lev}")
     # print(f"i... times  == {fev:.2f} ... {lev:.2f}")
-    dift = lev - fev
+    dift = lev - fev  # difference of times
     deadt = dift*deadtpr/100
     livet = dift - deadt
 
+    livetprsat = (dift - deadt - satu_DT_ch/1e+6)
+    deadtprsat = (dift - livetprsat)/dift*100
+
+
     stopcut = start + dt.timedelta(seconds=lev)
+    rate = len(df1)/dift
 
     output = f"""
      file   == {filename}
+    channel == {chan}
+     rate   == {rate:.1f} cps
      times  == {fev:.2f} ... {lev:.2f}
      real T == {dift:.2f} s
      live T == {livet:.2f} s
@@ -304,12 +432,33 @@ def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
      zeroes     = {len_dfzero:8d}
      nz events  = {len(df2):8d}
      tot events = {len(df1):8d}
+     saturations= {satu_DT_ch/1e+6:.2f} s
 
-     DT %       = {deadtpr:.2f}
+     DT %       = {deadtpr:.2f} %
+     DT(inc sat)= {deadtprsat:.2f} %
 """
     print(output)
 
-    outinfo = os.path.splitext(filename)[0]+".info"
+    df_stat['rate'] =  rate
+    df_stat['DT%'] = deadtpr
+    df_stat['DT+S%'] = deadtprsat
+
+    df_stat['realT'] =dift
+    df_stat['liveT'] = livet
+    df_stat['deadT'] = deadt
+    df_stat['Nzero'] =len_dfzero
+    df_stat['Ntot'] = len(df1)
+    df_stat['iDTavg'] = 100*( (1/rate) * satu_n_ch )/dift   #  average dt * nsatur
+    df_stat['iDTmin'] = 100*( df_stat['min_dtus']*1e-6 * satu_n_ch )/dift  #  mintime * nsatur
+#    df_stat['tSAT'] =
+
+#    df_stat = df_stat.round( decimals = 2 )
+    df_stat = df_stat.round( decimals = 2 )
+
+    print( tabulate(df_stat, headers='keys',showindex="never"))#, tablefmt='psql'
+    print()
+
+    outinfo = os.path.splitext(filename)[0]+f"_ch{chan}.info"
     print(f"D... creating info FILE {outinfo}")
     with open(outinfo,"w") as f:
         f.write(output)
@@ -321,7 +470,8 @@ def eva_cut_time(filename,  od=0, do=9999999 , chan=1,  tree = False):
     #*************
     hname = general.generate_hname(filename, chan)
 
-    outfile = bname+".txt"
+
+    outfile = bname+f"_ch{chan}.txt"
     #*************
     his = general.column_to_histo(df2['E'], savename = outfile, hname = hname, writeondisk = True)
     outfile2 = bname+".asc1"
